#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <popt.h>
#include <libharkio3.h>

#define ZDIR_TF          "transferFunction/"       //required
#define ZPATH_README     ZDIR_TF "README"          //optional
#define ZPATH_WHATISTHIS ZDIR_TF "whatisthis.txt"  //required
#define ZPATH_MIC        ZDIR_TF "microphones.xml" //required
#define ZPATH_SRC        ZDIR_TF "source.xml"      //required
#define ZDIR_TFLOC       ZDIR_TF "localization/"   //optional
#define ZDIR_TFSEP       ZDIR_TF "separation/"     //optional
#define ZPATH_TFLOC_PREF ZDIR_TFLOC "tf"           //either
#define ZPATH_TFSEP_PREF ZDIR_TFSEP "tf"           //either

#define ZFILENUM_REQUIRED  6
#define ZFILENUM_OPTIONAL  1

bool hasParsedOpt = false;
void help(poptContext con, int exitCode) {
    if (hasParsedOpt)
        return;

    printf("\n");
    printf("\n");
    printf("NAME:\n");
    printf("  harktoolcli-validtf\n");
    printf("\n");
    printf("Description:\n");
    printf("  Consistency check of the transfer function\n");
    printf("\n");
    poptPrintHelp(con, stdout, 0);
    printf("\n");
    printf("Output:\n");
    printf("  If a tf_file is valid, the command returns 0.\n");
    printf("\n");
    printf("\n");
    exit(exitCode);
}

// This function is a copy from libharkio3_transferfunction.c
int countTFs(struct zip *handle, char *mode, harkio_Positions *poses) {
    if (handle == NULL) {
        harkio_Log_printf(harkio_Log_Error, "zip file hanel is NULL");
        return -1;
    }

    harkio_Log_printf(harkio_Log_Debug, "size: %d", poses->size);
    if (poses->size < 0) {
        harkio_Log_printf(harkio_Log_Error, "maximum TF index is not positive value");
        return -1;
    }

    if (strncmp(mode, "localization", 12) == 0) {
        //intmode = MODE_LOC;
    } else if (strncmp(mode, "separation", 10) == 0) {
        //intmode = MODE_SEP;
    } else {
        harkio_Log_printf(harkio_Log_Error, "Unknown mode: %s", mode);
        return -1;
    }

    int i;
    for (i = 0; i < poses->size; i++) {
        char tmppath[4192];
        sprintf(tmppath, "transferFunction/%s/tf%05d.mat", mode, poses->pos[i]->id);
        harkio_Log_printf(harkio_Log_Debug, tmppath);
        if (zip_name_locate(handle, tmppath, 0) == -1)
            break;
    }
    return i;
}

// This function is a copy from libharkio3_transferfunction.c
harkio_XML* readZippedXML(struct zip *handle, char *path) {
    harkio_Log_printf(harkio_Log_Trace, "@ %s", __func__);

    char *buffer;
    struct zip_file *fp;
    struct zip_stat zs;
    harkio_XML *xml;

    //--------------------------------------------//
    // Argument Check                             //
    //--------------------------------------------//
    if (path == NULL) {
        harkio_Log_printf(harkio_Log_Error, "path is NULL");
        return NULL;
    }

    //--------------------------------------------//
    // main                                       //
    //--------------------------------------------//
    if ((fp = zip_fopen(handle, path, 0)) == NULL) {
        harkio_Log_FileOpenFailed(path);
        return NULL;
    }
    zip_stat_init(&zs);
    zip_stat(handle, path, ZIP_STAT_SIZE, &zs);

    buffer = (char *) malloc(zs.size + 1);
    zip_fread(fp, buffer, zs.size);
    buffer[zs.size] = '\0';

    xml = harkio_XML_newFromXMLbuffer(buffer, zs.size, NULL);

    free(buffer);
    zip_fclose(fp);

    return xml;
}

// This function is a copy from libharkio3_transferfunction.c
char* readZippedText(struct zip *handle, char *path) {
    harkio_Log_printf(harkio_Log_Trace, "@ %s", __func__);

    char *buffer;
    struct zip_file *fp;
    struct zip_stat zs;

    //--------------------------------------------//
    // Argument Check                             //
    //--------------------------------------------//
    if (path == NULL) {
        harkio_Log_printf(harkio_Log_Error, "path is NULL");
        return NULL;
    }

    //--------------------------------------------//
    // main                                       //
    //--------------------------------------------//
    if ((fp = zip_fopen(handle, path, 0)) == NULL) {
        harkio_Log_FileOpenFailed(path);
        return NULL;
    }
    zip_stat_init(&zs);
    zip_stat(handle, path, ZIP_STAT_SIZE, &zs);
    buffer = (char *) malloc(zs.size + 1);
    zip_fread(fp, buffer, zs.size);
    buffer[zs.size] = '\0';

    zip_fclose(fp);

    return buffer;
}

// This function is a copy from libharkio3_transferfunction.c and deleting harkio_Matrix_fromZip
harkio_TransferFunction* harkio_TransferFunction_fromFile_withoutLoadingMatrix(const char *path, char *signature) {
    harkio_Log_printf(harkio_Log_Trace, "@ %s", __func__);
    char *sign = "transfer function";

    int numLocTFs = 0;
    int numSepTFs = 0;

    //--------------------------------------------//
    // Argument Check                             //
    //--------------------------------------------//
    if (path == NULL) {
        harkio_Log_printf(harkio_Log_Error, "path is NULL");
        return NULL;
    }

    if (signature != NULL) {
        sign = signature;
    }
    harkio_Log_printf(harkio_Log_Debug, "whatisthis: %s", sign);

    //--------------------------------------------//
    // main                                       //
    //--------------------------------------------//
    // harkio_TransferFunction *tf = (harkio_TransferFunction*) calloc(sizeof(harkio_TransferFunction), 1);
    harkio_TransferFunction *tf = harkio_TransferFunction_new();

    if (tf == NULL) {
        harkio_Log_printf(harkio_Log_Error, "can not allocated memory");
        harkio_TransferFunction_delete(&tf);
        return NULL;
    }

    harkio_XML *xml;
    int errorp;
    //int i;
    char *buffer;

    if ((tf->handle = zip_open(path, ZIP_CHECKCONS, &errorp)) == NULL) {
        if (errorp == ZIP_ER_OPEN) { // NO FILE
            harkio_Log_printf(harkio_Log_Error, "zip_open failed: cannot find or open '%s'", path);
            return NULL;

        } else if (errorp == ZIP_ER_NOZIP) { // NOT IN ZIP FORMAT
            harkio_Log_printf(harkio_Log_Error, "zip_open failed: file '%s' is not in zip format", path);

            // COMPAT HARK FILE CHECK
            FILE* fpHARK = NULL;
            if ((fpHARK = fopen(path, "rb")) != NULL) {
                char head[4] = { '\0' };
                int fread_ret = fread(head, 1, 4, fpHARK);
                fclose(fpHARK);
                if (fread_ret == 4 && strncmp(head, "HARK", 4) == 0) {
                    harkio_Log_printf(harkio_Log_Error, "\n"
                            "   If you want to convert your binary-format TFs into a zip-format (HARK>=2.1),\n"
                            "   see 'harktoolcli-conv-tf --help' (after installing harktool5).\n");
                    harkio_TransferFunction_delete(&tf);
                    return NULL;
                }
            }

        } else {
            harkio_Log_printf(harkio_Log_Error, "zip_open failed: file '%s' unexpected error", path);
        }
        harkio_TransferFunction_delete(&tf);
        return NULL;
    }

    // if transferFunction/whatisthis.txt != transfer function, return NULL
    buffer = readZippedText(tf->handle, "transferFunction/whatisthis.txt");
    if (buffer == NULL) {
        harkio_Log_printf(harkio_Log_Error, "readZippedText failed");
        harkio_TransferFunction_delete(&tf);
        return NULL;
    }
    harkio_Log_printf(harkio_Log_Debug, buffer);

    if (strncmp(buffer, sign, strlen(sign)) != 0) {
        char tmp[4192];
        sprintf(tmp, "[transferFunction/whatisthis.txt] is not [%s]", sign);
        harkio_Log_printf(harkio_Log_Warning, tmp);
        harkio_TransferFunction_delete(&tf);
        free(buffer);
        return NULL;
    }

    free(buffer);

    // read sources.xml
    harkio_Log_printf(harkio_Log_Debug, "reading sources.xml");
    xml = readZippedXML(tf->handle, "transferFunction/source.xml");
    if (xml == NULL) {
        harkio_Log_printf(harkio_Log_Error, "source.xml is not found in the file");
        harkio_TransferFunction_delete(&tf);
        return NULL;
    }
    if (xml->poses) {
        harkio_Log_printf(harkio_Log_Debug, "copying xml->poses");
        // tf->poses = harkio_Positions_copy(xml->poses);
        harkio_TransferFunction_setPositions(tf, harkio_Positions_copy(xml->poses));
    }

    if (xml->cfg) {
        harkio_Log_printf(harkio_Log_Debug, "copying xml->cfg");
        // tf->cfg   = harkio_Config_copy(xml->cfg);
        harkio_TransferFunction_setConfig(tf, harkio_Config_copy(xml->cfg));
    }

    if (xml->nbrs) {
        harkio_Log_printf(harkio_Log_Debug, "copying xml->nbrs");
        // tf->nbrs  = harkio_Neighbors_copy(xml->nbrs);
        harkio_TransferFunction_setNeighbors(tf, harkio_Neighbors_copy(xml->nbrs));
    }

    harkio_Log_printf(harkio_Log_Debug, "reading sources.xml");
    if (tf->nbrs != NULL)
        tf->nbrs->poses = tf->poses;
    harkio_XML_delete(&xml);

    // read microphones.xml
    harkio_Log_printf(harkio_Log_Debug, "reading microphones.xml");
    xml = readZippedXML(tf->handle, "transferFunction/microphones.xml");
    if (xml) {
        harkio_TransferFunction_setMicrophones(tf, harkio_Positions_copy(xml->poses));
        // tf->mics = harkio_Positions_copy(xml->poses);
    } else {
        tf->mics = NULL;
    }

    harkio_XML_delete(&xml);

    // read localization/separation tfs
    harkio_Log_printf(harkio_Log_Debug, "reading loc/sep tfs");
    tf->loctfs = tf->septfs = NULL;
    numLocTFs = countTFs(tf->handle, "localization", tf->poses);
    numSepTFs = countTFs(tf->handle, "separation", tf->poses);

    if (numLocTFs < 0) {
        harkio_Log_printf(harkio_Log_Error, "Reading localization TFs failed.");
        return NULL;
    }

    if (numSepTFs < 0) {
        harkio_Log_printf(harkio_Log_Error, "Reading separation TFs failed.");
        return NULL;
    }

    if (numLocTFs > 0) {
        //tf->loctfs = (harkio_Matrix **) malloc(sizeof(harkio_Matrix *) * tf->poses->size);
        harkio_TransferFunction_setLocTFs(tf, NULL);
    }
    if (numSepTFs > 0) {
        // tf->septfs = (harkio_Matrix **) malloc(sizeof(harkio_Matrix *) * tf->poses->size);
        harkio_TransferFunction_setSepTFs(tf, NULL);
    }

    /*for (i = 0; i < tf->poses->size; i++) {
        char tmppath[4192];

        if (numLocTFs > 0) {
            sprintf(tmppath, "transferFunction/localization/tf%05d.mat", tf->poses->pos[i]->id);
            // tf->loctfs[i] = harkio_Matrix_fromZip(tf->handle, tmppath);
            harkio_Matrix* mat = harkio_Matrix_fromZip(tf->handle, tmppath);
            harkio_TransferFunction_appendLocTF(tf, mat);
        }
        if (numSepTFs > 0) {
            sprintf(tmppath, "transferFunction/separation/tf%05d.mat", tf->poses->pos[i]->id);
            // tf->septfs[i] = harkio_Matrix_fromZip(tf->handle, tmppath);
            harkio_Matrix* mat = harkio_Matrix_fromZip(tf->handle, tmppath);
            harkio_TransferFunction_appendSepTF(tf, mat);
        }
    }*/
    tf->internal_locsize = numLocTFs;
    tf->internal_sepsize = numSepTFs;

    return tf;
}


int main(int argc, char **argv) {

    int errorp;
    struct zip *handle;

    int posCnt;
    int isShow = false;
    int isJson = false;
    char* logLevel = "E";

    struct poptOption optionsTable[] = {
    /*{ "longname", "shortname",
     *  argInfo,
     *  *arg, retval,
     *  description, argment description} */
    { "show", 's',
      POPT_ARG_NONE,
      &isShow, 's',
      "Show details", NULL },
    { "json", 'j',
      POPT_ARG_NONE | POPT_ARGFLAG_DOC_HIDDEN,
      &isJson, 'j',
      "Show details (JSON)", NULL },
    { "logLevel", '\0',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &logLevel, '-',
      "Log information level", "E|W|I|D" },
    { "help", '\0',
      0,
      NULL, '?',
      "Show this help message", NULL },
    POPT_TABLEEND };

    poptContext optCon = poptGetContext(NULL, argc, (const char**)argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "[OPTIONS]* tf_filename(zip)");

    // argc check
    if (argc <= 1) {
        harkio_Log_printf(harkio_Log_Error, "argument error!!");
        help(optCon, EXIT_FAILURE);
    }

    // option processing
    int c;
    while ((c = poptGetNextOpt(optCon)) >= 0) {
        switch (c) {
        case '-':
            if (strcmp(logLevel, "W") == 0) {
                harkio_Log_setLevel(harkio_Log_Warning);
            } else if (strcmp(logLevel, "I") == 0) {
                harkio_Log_setLevel(harkio_Log_Info);
            } else if (strcmp(logLevel, "D") == 0) {
                harkio_Log_setLevel(harkio_Log_Debug);
            } else {
                harkio_Log_setLevel(harkio_Log_Error);
            }
            hasParsedOpt = true;
            break;
        case '?':
            // help
            help(optCon, EXIT_SUCCESS);
            /* no break */
        default:
            hasParsedOpt = true;
        }
    }
    if (c < -1) {
        /* an error occurred during option processing */
        harkio_Log_printf(harkio_Log_Error, "%s: %s",
                poptBadOption(optCon, POPT_BADOPTION_NOALIAS), poptStrerror(c));
        exit(EXIT_FAILURE);
    }

    harkio_Log_printf(harkio_Log_Debug, "-s=%s", (isShow) ? "true" : "false");

    // get filename
    const char *filename = poptGetArg(optCon);

    if (!filename) {
        harkio_Log_printf(harkio_Log_Error , "missing tf_filename");
        exit(EXIT_FAILURE);
    }

    // validate tf file
    handle = zip_open(filename, ZIP_CHECKCONS, &errorp);

    if (handle == NULL) {
        harkio_Log_printf(harkio_Log_Error, "can not open zip file [%s]", filename);
        exit(EXIT_FAILURE);
    }

    bool isTFValid = true;
    if (zip_name_locate(handle, ZDIR_TF, 0) == -1) {
        harkio_Log_printf(harkio_Log_Error , "%s not found.", ZDIR_TF);
        isTFValid = false;
    }
    if (zip_name_locate(handle, ZPATH_WHATISTHIS, 0) == -1) {
        harkio_Log_printf(harkio_Log_Error , "%s not found.", ZPATH_WHATISTHIS);
        isTFValid = false;
    }
    if (zip_name_locate(handle, ZPATH_MIC, 0) == -1) {
        harkio_Log_printf(harkio_Log_Error , "%s not found.", ZPATH_MIC);
        isTFValid = false;
    }
    if (zip_name_locate(handle, ZPATH_SRC, 0) == -1) {
        harkio_Log_printf(harkio_Log_Error , "%s not found.", ZPATH_SRC);
        isTFValid = false;
    }

    zip_close(handle);
    handle = NULL;

    if (!isTFValid) {
        return (EXIT_FAILURE);
    }

    //get position size
    harkio_TransferFunction *tf = NULL;

    tf = harkio_TransferFunction_fromFile_withoutLoadingMatrix(filename, NULL);
    if (tf == NULL) {
        harkio_Log_printf(harkio_Log_Error, "zip read file error");
        exit(EXIT_FAILURE);
    }
    posCnt = tf->poses->size;

    if (isShow) {
        fprintf(stdout, "Transfer Function Informations\n");
        // number of mics
        fprintf(stdout, " Microphones       : %d\n", tf->mics->size);
        // number of tf directions
        fprintf(stdout, " TF Directions     : %d\n", tf->poses->size);
        // contents of tf
        char* contents;
        if (tf->internal_locsize > 0 && tf->internal_sepsize > 0)
            contents = "localization, separation";
        else if (tf->internal_locsize > 0)
            contents = "localization";
        else if (tf->internal_sepsize > 0)
            contents = "separation";
        else
            contents = "no tf matrix";
        fprintf(stdout, " Contents          : %s\n", contents);

        fprintf(stdout, "Configuration Informations\n");
        // config:comment
        fprintf(stdout, " Comment           : %s\n",
                (tf->cfg->comment) ? tf->cfg->comment : "");
        // SynchronousAverage
        fprintf(stdout, " SynchronousAverage: %d\n",
                tf->cfg->synchronousAverage);
        // TSPpath
        fprintf(stdout, " TSPpath           : %s\n",
                (tf->cfg->TSPpath) ? tf->cfg->TSPpath : "");
        // TSPOffset
        fprintf(stdout, " TSPOffset         : %d\n", tf->cfg->TSPoffset);
        // PeakSearch from to
        fprintf(stdout, " PeakSearch        : from %d to %d\n",
                tf->cfg->peakSearchFrom, tf->cfg->peakSearchTo);
        // nfft
        fprintf(stdout, " nfft              : %d\n", tf->cfg->nfft);
        // samplingRate
        fprintf(stdout, " samplingRate      : %d\n", tf->cfg->samplingRate);
        // signalMax
        fprintf(stdout, " signalMax         : %d\n", tf->cfg->signalMax);
        // TSPLength
        fprintf(stdout, " TSPLength         : %d\n", tf->cfg->TSPLength);
    } else if (isJson) {
        fprintf(stdout, "{\n    \"TFInfo\": {\n");
        // number of mics
        fprintf(stdout, "        \"Microphones\": %d ,\n", tf->mics->size);
        // number of tf directions
        fprintf(stdout, "        \"Directions\": %d ,\n", tf->poses->size);
        // contents of tf
        char* contents;
        if (tf->internal_locsize > 0 && tf->internal_sepsize > 0)
            contents = "[ \"localization\" , \"separation\" ]";
        else if (tf->internal_locsize > 0)
            contents = "[ \"localization\" ]";
        else if (tf->internal_sepsize > 0)
            contents = "[ \"separation\" ]";
        else
            contents = "[]";
        fprintf(stdout, "        \"Contents\": %s\n    } ,\n", contents);

        fprintf(stdout, "    \"Config\": {\n");
        // config:comment
        fprintf(stdout, "        \"Comment\": \"%s\" ,\n",
                (tf->cfg->comment) ? tf->cfg->comment : "");
        // SynchronousAverage
        fprintf(stdout, "        \"SynchronousAverage\": %d ,\n",
                tf->cfg->synchronousAverage);
        // TSPpath
        fprintf(stdout, "        \"TSPpath\": \"%s\" ,\n",
                (tf->cfg->TSPpath) ? tf->cfg->TSPpath : "");
        // TSPOffset
        fprintf(stdout, "        \"TSPOffset\": %d ,\n", tf->cfg->TSPoffset);
        // PeakSearch from to
        fprintf(stdout, "        \"PeakSearch\": [ { \"from\" : %d, \"to\" : %d } ] ,\n",
                tf->cfg->peakSearchFrom, tf->cfg->peakSearchTo);
        // nfft
        fprintf(stdout, "        \"nfft\": %d ,\n", tf->cfg->nfft);
        // samplingRate
        fprintf(stdout, "        \"samplingRate\": %d ,\n", tf->cfg->samplingRate);
        // signalMax
        fprintf(stdout, "        \"signalMax\": %d ,\n", tf->cfg->signalMax);
        // TSPLength
        fprintf(stdout, "        \"TSPLength\": %d\n    }\n}", tf->cfg->TSPLength);
    }

    if (tf->internal_locsize > 0 && tf->internal_locsize != posCnt) {
        harkio_Log_printf(harkio_Log_Error, "localization matrix size(%d) differ from position size(%d)", tf->internal_locsize, posCnt);
        isTFValid = false;
    }
    if (tf->internal_sepsize > 0 && tf->internal_sepsize != posCnt) {
        harkio_Log_printf(harkio_Log_Error, "separation matrix size(%d) differ from position size(%d)", tf->internal_sepsize, posCnt);
        isTFValid = false;
    }

    harkio_TransferFunction_delete(&tf);

    // free poptContext
    poptFreeContext(optCon);

    if (isTFValid) {
        fprintf(stderr, "SUCCESS\n"); // "Successfully validated."
        return EXIT_SUCCESS;
    } else {
        return EXIT_FAILURE;
    }
}

